package algorithms.Method2RMadness;

import java.util.*;

import utils.ClassHierarchy;
import utils.MethodID;

import de.fub.bytecode.classfile.*;
import de.fub.bytecode.generic.*;




/** Shuffles the argument order of a method.
  *
  * <p>
  * Used by Method2RMadness obfuscator.
  *
  * @author Martin Stepp (<a href="mailto:stepp@cs.arizona.edu">stepp@cs.arizona.edu</a>)
  */
public class ParameterReorderer   implements de.fub.bytecode.Constants
{
        /* Turn to true to print debugging messages. */
        private static final boolean DEBUG = false;
        private static final boolean DEBUG2 = false;

        /* Turn to true to print duplicate method debugging messages. */
        private static final boolean DUP_DEBUG = false;

        private static final int LOAD = 0, STORE = 1, INC = 2;

        /* Required for SandMark's interface. */
        private utils.ConfigProperties myProps = new utils.ConfigProperties();


        /* Makes an ICONST_n or BIPUSH n for a given int. */
        private static Instruction createPushInstruction(int n)
        {
                switch (n)
                {
                        case 0:
                        case 1:
                        case 2:
                        case 3:
                        case 4:
                        case 5:
                                return new ICONST(n);
                        default:
                                return new BIPUSH((byte)n);
                }
        }


        /* Runs a quick test of this obfuscator. */
        public static void main(String[] args) throws Exception
        {
                if (args.length < 1)
                {
                        System.out.println("Usage: ParameterReorder JAR_FILE");
                        System.exit(1);
                }

                try
                {
                        utils.ClassFileCollection cfc = new utils.ClassFileCollection(args[0]);
                        ParameterReorderer reorderer = new ParameterReorderer();
                        reorderer.apply(cfc);

                        System.out.println("\nCFC after: ");
                        
                        cfc.saveJar("PARAM_HARDCODE.jar");
                }
                catch (java.io.IOException e)
                {
                        System.err.println("I/O Error: " + e.getMessage());
                }
        }


        /* To be implemented; returns whether or not we're allowed to reorder the arguments in this method. */
        private static boolean canReorder(MethodGen methGen)
        {
                if (methGen == null)
                        return false;

                // only operate on static methods
                if (!methGen.isStatic()  ||  methGen.isAbstract()  ||  methGen.isNative()  ||  methGen.isVolatile())
                        return false;

                // kludge
                String methodName = methGen.getName();
                if (methodName.equals("main")  ||  methodName.equals("<clinit>"))
                {
                        if (DEBUG)  System.out.println("special method main or class initializer!  not renaming.");
                        return false;
                }

                // make sure method's args are all reference types (Objects, not primitives)
                Type[] varTypes = methGen.getArgumentTypes();
                for (int i = 0;  i < varTypes.length;  i++)
                {
                        if (!(varTypes[i] instanceof ReferenceType))
                                return false;
                }

                return true;
        }

        private static boolean canReorder(JavaClass jc, ClassHierarchy ch,
                        de.fub.bytecode.generic.MethodGen methGen, String newSig)
        {
                if (!canReorder(methGen))
                        return false;

                // only operate on appropriate classes
                if (jc.isInterface())
                        return false;

                // don't rename methods that Sandmark flags as being non-renamable
                String className  = jc.getClassName();
                String methodName = methGen.getName();
                String methodSig  = methGen.getSignature();
                boolean isStatic  = methGen.isStatic();

                MethodID methodID   = new utils.MethodID(methodName, methodSig, className);;
                MethodID renameToID = new utils.MethodID(methodName, newSig, className);

                boolean okToRename = true;
                int numOverloads = -1;
                try {
                   okToRename = ch.methodRenameOK(methodID, renameToID);
                   if (okToRename)
                   {
                        MethodID[] overloads = ch.allOverloads(methodID);
                        if (overloads != null)
                                numOverloads = overloads.length;
                   }
                }catch(utils.ClassHierarchyException exc) {
                   // CC
                }
                if (DEBUG)
                {
                        System.out.println("old method: " + methodID);
                        System.out.println("new method: " + renameToID);
                        System.out.println("rename ok?: " + okToRename);
                        System.out.println("overloads : " + numOverloads);
                }

                // only allow Sandmark-approved renamable methods or ones that have no
                // inheritance overloading tree associated with them
                return okToRename  &&  numOverloads <= 1;
        }


        public static void printCFC(utils.ClassFileCollection cfc)
                throws Exception
        {
                java.util.Iterator itr = cfc.classes();
                while (itr.hasNext())
                {
                        String className = (String)itr.next();
                        System.out.println("Class " + className);
                        de.fub.bytecode.classfile.JavaClass jc = cfc.getClass(className);

                        de.fub.bytecode.generic.ClassGen cgen = new de.fub.bytecode.generic.ClassGen(jc);
                        de.fub.bytecode.generic.ConstantPoolGen cpgen = cgen.getConstantPool();

                        // shuffle all methods in this class
                        int numMethods = cgen.getMethods().length;
                        for (int methodNum = 0;  methodNum < numMethods;  methodNum++)
                        {
                                // reorder each method
                                de.fub.bytecode.classfile.Method meth = cgen.getMethodAt(methodNum);
                                de.fub.bytecode.generic.MethodGen methGen  = new de.fub.bytecode.generic.MethodGen(meth, className, cpgen);
                                System.out.println("  Method " + meth);
                                // System.out.println(methGen.getInstructionList() + "\n");
                        }
                }
        }

        private static String getMethodKey(JavaClass jc, Method meth)
        {
                return getMethodKey(jc.getClassName(), meth.getName(), meth.getSignature());
        }

        private static String getMethodKey(String className,
                String methodName, String methodSignature)
        {
                return className + ":" + methodName + methodSignature;
        }

        /** Reorders the method arguments in this ClassFileCollection's classes. */
        public void apply(utils.ClassFileCollection cfc)
                throws Exception
        {
                // create a class hierarchy for later seeing if methods are renamable
                utils.ClassHierarchy ch = new utils.ClassHierarchy(cfc, myProps);

                java.util.ArrayList classNameList = new java.util.ArrayList();
                java.util.ArrayList classList = new java.util.ArrayList();

                java.util.Iterator itr = cfc.classes();
                while (itr.hasNext())
                        classNameList.add((String)itr.next());

                Hashtable varIndexTable    = new Hashtable();
                Hashtable newIndexTable    = new Hashtable();
                Hashtable methodGenTable   = new Hashtable();
                Hashtable oldArgTypesTable = new Hashtable();
                Hashtable newArgTypesTable = new Hashtable();
                Hashtable newMethodTable   = new Hashtable();

                int numClasses = classNameList.size();
                for (int classNum = 0;  classNum < numClasses;  classNum++)
                {
                        String className = (String)classNameList.get(classNum);
                        JavaClass jc = cfc.getClass(className);

                        ClassGen cgen = new ClassGen(jc);
                        ConstantPoolGen cpgen = cgen.getConstantPool();

                        // shuffle all methods in this class
                        int numMethods = cgen.getMethods().length;
                        for (int methodNum = 0;  methodNum < numMethods;  methodNum++)
                        {
                                                                // reorder each method
                                de.fub.bytecode.classfile.Method meth = cgen.getMethodAt(methodNum);
                                if (DEBUG)  System.out.println("Old method: " + meth);

                                MethodGen methGen  = new MethodGen(meth, className, cpgen);

                                if (!canReorder(methGen))
                                {
                                        if (DEBUG)  System.out.println("Can't reorder(1): " + meth);
                                        continue;
                                }

                                // randomize numbers 0--n, where n is number of arguments
                                String[] oldArgNames = methGen.getArgumentNames();
                                Type[] oldArgTypes   = methGen.getArgumentTypes();

                                int numArgs = oldArgNames.length;
                                        // don't rearrange methods with less than 2 args (nothing to do!)
                                if (numArgs <= 1)
                                {
                                        continue;
                                }

                                // if (DEBUG)  System.out.println(methGen.getInstructionList() + "\n");

                                // try again; shuffle args 0--n-1
                                ArrayList argNums = new ArrayList();
                                Hashtable varNumTable = new Hashtable();
                                for (int ii = 0; ii < numArgs; ii++)
                                        argNums.add(new Integer(ii));

                                Collections.shuffle(argNums);

                                String[] newArgNames = new String[numArgs];
                                Type[] newArgTypes = new Type[numArgs];
                                int[] indeces = new int[numArgs];
                                int[] newIndeces = new int[numArgs];
                                int numChanged = 0;
                                for (int oldIndex = 0; oldIndex < numArgs; oldIndex++)
                                {
                                        Integer inewIndex = (Integer)argNums.get(oldIndex);
                                        int newIndex = inewIndex.intValue();
                                        if (oldIndex != newIndex)
                                                numChanged++;

                                        // remap the arguments by their new indeces
                                        indeces[newIndex] = oldIndex;
                                        newIndeces[oldIndex] = newIndex;
                                        newArgNames[newIndex] = oldArgNames[oldIndex];
                                        newArgTypes[newIndex] = oldArgTypes[oldIndex];

                                        // hash in [old arg #] ==> [new arg #] table
                                        Integer key = new Integer(oldIndex);
                                        Integer value = new Integer(newIndex);
                                        varNumTable.put(key, value);
                                }

                                if (numChanged == 0)
                                {
                                        if (DEBUG)  System.out.println("No arguments changed!  Moving to next method.");
                                        continue;
                                }


                                // apply the changes to the arg names/types
                                methGen.setArgumentNames(newArgNames);
                                methGen.setArgumentTypes(newArgTypes);

                                // make sure we can reorder this method
                                String newSig = methGen.getSignature();
                                if (!canReorder(jc, ch, methGen, newSig))
                                {
                                        if (DEBUG)  System.out.println("Can't reorder (2): " + meth);
                                        continue;
                                }

                                // hash in [method key] => [old method Type[]] table
                                // and others
                                String methodKey = getMethodKey(jc, meth);
                                oldArgTypesTable.put(methodKey, oldArgTypes);
                                newArgTypesTable.put(methodKey, newArgTypes);
                                newMethodTable.put(methodKey, methGen);

/*
                                // not needed any more -- assumes all ref type
                                LocalVariableGen[] new_lvars = methGen.getLocalVariables();
                                for (int i = 0;  i < new_lvars.length;  i++)
                                {
                                        System.out.println("NEW LV #" + i + " index is " + new_lvars[i].getIndex());
                                }
*/

                                // if (DEBUG2)  System.out.println("varNumTable = " + varNumTable + "\n");
                                // if (DEBUG)  System.out.println("arrayIndexTable = " + arrayIndexTable + "\n");

                                // fix instructions so they refer to the now-rearranged local variable arguments
                                // this involves updating any local variable instruction (a load, store, or int increment)
                                // to make it use the new local variable numbers and types.
                                if (DEBUG)  System.out.println("patching instructions in method:\n---");

                                InstructionList instrList = methGen.getInstructionList();
                                for (InstructionHandle ihandle = instrList.getStart();
                                     ihandle != null;
                                     ihandle = ihandle.getNext())
                                {
                                        Instruction instr = ihandle.getInstruction();
                                        // de.fub.bytecode.generic.Instruction instr = instrs[instrNum];
                                        if (!(instr instanceof de.fub.bytecode.generic.LocalVariableInstruction))
                                                // we only have to modify instructions that refer to local vars (arguments)
                                                continue;

                                        LocalVariableInstruction lvinstr = (LocalVariableInstruction)instr;
                                        // if (DEBUG)  System.out.println("L.V. instr: " + instr);
                                        Type oldType = null;
                                        int oldNum = 0;
                                        int opcode = instr.getOpcode();
                                        int incAmount = 0;
                                        int instrFamily = -1;

                                        // figure out what number of argument this instruction referred to
                                        switch (opcode)
                                        {
                                                case ALOAD_0:  case DLOAD_0:  case FLOAD_0:  case ILOAD_0:  case LLOAD_0:
                                                case ASTORE_0: case DSTORE_0: case FSTORE_0: case ISTORE_0: case LSTORE_0:
                                                        oldNum = 0;
                                                        break;
                                                case ALOAD_1:  case DLOAD_1:  case FLOAD_1:  case ILOAD_1:  case LLOAD_1:
                                                case ASTORE_1: case DSTORE_1: case FSTORE_1: case ISTORE_1: case LSTORE_1:
                                                        oldNum = 1;
                                                        break;
                                                case ALOAD_2:  case DLOAD_2:  case FLOAD_2:  case ILOAD_2:  case LLOAD_2:
                                                case ASTORE_2: case DSTORE_2: case FSTORE_2: case ISTORE_2: case LSTORE_2:
                                                        oldNum = 2;
                                                        break;
                                                case ALOAD_3:  case DLOAD_3:  case FLOAD_3:  case ILOAD_3:  case LLOAD_3:
                                                case ASTORE_3: case DSTORE_3: case FSTORE_3: case ISTORE_3: case LSTORE_3:
                                                        oldNum = 3;
                                                        break;
                                                case ALOAD:        case DLOAD:    case FLOAD:    case ILOAD:    case LLOAD:
                                                case ASTORE:   case DSTORE:   case FSTORE:   case ISTORE:   case LSTORE:
                                                        oldNum = lvinstr.getIndex();
                                                        break;
                                                case IINC:
                                                        oldNum = lvinstr.getIndex();
                                                        incAmount = ((IINC)lvinstr).getIncrement();
                                                        break;
                                        }

                                        if (oldNum >= numArgs)
                                        {
                                                // this instruction refers to a non-argument local variable;
                                                // we don't need to modify it
                                                // if (DEBUG)  System.out.println("  Skipping; number out of range: " + oldNum);
                                                continue;
                                        }

                                        Integer inewNum = (Integer)varNumTable.get(new Integer(oldNum));
                                        if (inewNum == null)
                                        {
                                                // if (DEBUG)  System.out.println("  inewNum null!  No varNumTable entry for this variable.");
                                                continue;
                                        }

                                        int newNum = inewNum.intValue();
                                        if (oldNum == newNum)
                                        {
                                                // don't need to modify instruction; it's the same index as it used to be
                                                // if (DEBUG)  System.out.println("  Skipping; old num same as new num for LV " + oldNum);
                                                continue;
                                        }

                                        switch (opcode)
                                        {
                                                case ALOAD_0:  case DLOAD_0:  case FLOAD_0:  case ILOAD_0:  case LLOAD_0:
                                                case ALOAD_1:  case DLOAD_1:  case FLOAD_1:  case ILOAD_1:  case LLOAD_1:
                                                case ALOAD_2:  case DLOAD_2:  case FLOAD_2:  case ILOAD_2:  case LLOAD_2:
                                                case ALOAD_3:  case DLOAD_3:  case FLOAD_3:  case ILOAD_3:  case LLOAD_3:
                                                case ALOAD:    case DLOAD:    case FLOAD:    case ILOAD:    case LLOAD:
                                                        instrFamily = LOAD;
                                                        break;
                                                case ASTORE_0: case DSTORE_0: case FSTORE_0: case ISTORE_0: case LSTORE_0:
                                                case ASTORE_1: case DSTORE_1: case FSTORE_1: case ISTORE_1: case LSTORE_1:
                                                case ASTORE_2: case DSTORE_2: case FSTORE_2: case ISTORE_2: case LSTORE_2:
                                                case ASTORE_3: case DSTORE_3: case FSTORE_3: case ISTORE_3: case LSTORE_3:
                                                case ASTORE:   case DSTORE:   case FSTORE:   case ISTORE:   case LSTORE:
                                                        instrFamily = STORE;
                                                        break;
                                                case IINC:
                                                        instrFamily = INC;
                                                        break;
                                        }

                                        // figure out what type this instruction referred to
                                        switch (opcode)
                                        {
                                                case ALOAD_0:  case ALOAD_1:  case ALOAD_2:  case ALOAD_3:  case ALOAD:
                                                case ASTORE_0: case ASTORE_1: case ASTORE_2: case ASTORE_3: case ASTORE:
                                                        oldType = Type.OBJECT;
                                                        break;
                                                case DLOAD_0:  case DLOAD_1:  case DLOAD_2:  case DLOAD_3:  case DLOAD:
                                                case DSTORE_0: case DSTORE_1: case DSTORE_2: case DSTORE_3: case DSTORE:
                                                        oldType = Type.DOUBLE;
                                                        break;
                                                case FLOAD_0:  case FLOAD_1:  case FLOAD_2:  case FLOAD_3:  case FLOAD:
                                                case FSTORE_0: case FSTORE_1: case FSTORE_2: case FSTORE_3: case FSTORE:
                                                        oldType = Type.FLOAT;
                                                        break;
                                                case ILOAD_0:  case ILOAD_1:  case ILOAD_2:  case ILOAD_3:  case ILOAD:
                                                case ISTORE_0: case ISTORE_1: case ISTORE_2: case ISTORE_3: case ISTORE:
                                                case IINC:
                                                        oldType = Type.INT;
                                                        break;
                                                case LLOAD_0:  case LLOAD_1:  case LLOAD_2:  case LLOAD_3:  case LLOAD:
                                                case LSTORE_0: case LSTORE_1: case LSTORE_2: case LSTORE_3: case LSTORE:
                                                        oldType = Type.LONG;
                                                        break;
                                        }

                                        // build new instruction that refers correctly to new local variable
                                        Instruction newInstr = null;
                                        switch (instrFamily)
                                        {
                                                case LOAD:
                                                        newInstr = InstructionFactory.createLoad(oldType, newNum);
                                                        break;
                                                case STORE:
                                                        newInstr = InstructionFactory.createStore(oldType, newNum);
                                                        break;
                                                case INC:
                                                        newInstr = new IINC(newNum, incAmount);
                                                        break;
                                        }

                                        // if (DEBUG)  System.out.println("  old num = " + oldNum + ", old type = " + oldType);
                                        // if (DEBUG)  System.out.println("  new num = " + newNum);
                                        // if (DEBUG)  System.out.println("  new instr = " + newInstr);

                                        // update method's instruction list to replace old instruction with new one
                                        ihandle.setInstruction(newInstr);
                                }

                                methGen.setInstructionList(instrList);
                                de.fub.bytecode.classfile.Method newMeth = methGen.getMethod();

                                if (DEBUG)  System.out.println("\nNew method: " + newMeth);
                                // if (DEBUG)  System.out.println(methGen.getInstructionList() + "\n");

                                // replace method in class
                                if (DUP_DEBUG)  System.out.println("\nFIRST PHASE, REPLACE METHOD:\n" + meth + " ,\n" + newMeth);
                                cgen.replaceMethod(meth, newMeth);

                                // store the hash table of [old var #] ==> [new var #], so we can fix calls later
                                varIndexTable.put(methodKey, indeces);
                                newIndexTable.put(methodKey, newIndeces);

                                // hash in [old method signature] ==> [new method gen] table
                                methodGenTable.put(methodKey, methGen);
                        }

                        // done shuffling all methods in this class;
                        // now replace old class with new one
                        JavaClass newClass = cgen.getJavaClass();
                        classList.add(cgen);
                        if (DUP_DEBUG)  System.out.println("CFC ADD: " + jc.getClassName());
                        cfc.addClass(newClass);
                }


                // fix all calls by making them push in new order onto stack
                if (DEBUG)  System.out.println("\nfixing all calls to modified methods:\n---");

                for (int classNum = 0;  classNum < numClasses;  classNum++)
                {
                        String callingClassName = (String)classNameList.get(classNum);
//                      JavaClass jc = (JavaClass)classList.get(classNum);

                        ClassGen cgen = (ClassGen)classList.get(classNum);  // new ClassGen(jc);
                        ConstantPoolGen cpgen = cgen.getConstantPool();
                        InstructionFactory ifactory = new InstructionFactory(cgen, cpgen);

                        boolean modified = false;

                        Method[] methods = cgen.getMethods();
                        int numMethods = methods.length;
                        for (int methodNum = 0;  methodNum < numMethods;  methodNum++)
                        {
                                de.fub.bytecode.classfile.Method meth = methods[methodNum];
                                de.fub.bytecode.generic.MethodGen methGen  = new de.fub.bytecode.generic.MethodGen(meth, callingClassName, cpgen);
                                InstructionList instrList = methGen.getInstructionList();
                                if (instrList == null)
                                        // empty method!
                                        continue;

                                for (InstructionHandle ihandle = instrList.getStart();
                                     ihandle != null;
                                     ihandle = ihandle.getNext())
                                {
                                        // only modify INVOKESTATIC instructions
                                        Instruction instr = ihandle.getInstruction();
                                        if (!(instr instanceof INVOKESTATIC))
                                                continue;

                                        InvokeInstruction invoke_instr = (InvokeInstruction)instr;
                                        if (DEBUG)  System.out.println("in " + callingClassName + ":" + meth + " at " + instr + ":");
                                        // if (DEBUG)  System.out.println("  methodName = " + methodName);
                                        // if (DEBUG)  System.out.println("  cp[i] = " + cpgen.getConstant(invoke_instr.getIndex()));

                                        String className  = invoke_instr.getClassName(cpgen);
                                        String methodName = invoke_instr.getMethodName(cpgen);
                                        String methodSig  = invoke_instr.getSignature(cpgen);
/*
                                        ConstantMethodref cMethodref = (ConstantMethodref)cpgen.getConstant(invoke_instr.getIndex());
                                        ConstantClass cClass = (ConstantClass)cpgen.getConstant(cMethodref.getClassIndex());
                                        ConstantUtf8 cClassName = (ConstantUtf8)cpgen.getConstant(cClass.getNameIndex());

                                        String className = cClassName.getBytes();
                                        ConstantNameAndType cNameAndType = (ConstantNameAndType)cpgen.getConstant(cMethodref.getNameAndTypeIndex());
                                        String methodName = cNameAndType.getName(cpgen.getConstantPool());
                                        String methodSig  = cNameAndType.getSignature(cpgen.getConstantPool());
*/
                                        if (DEBUG)  System.out.println("INVOKESTATIC " + className + "." + methodName + methodSig);

                                        String methodKey = getMethodKey(className, methodName, methodSig);
                                        if (!varIndexTable.containsKey(methodKey))
                                        {
                                                if (DEBUG)  System.out.println("  Skipping; can't find variable reordering table!");
                                                continue;
                                        }

                                        int[] indeces = (int[])varIndexTable.get(methodKey);
                                        int[] newIndeces = (int[])newIndexTable.get(methodKey);
                                        // int offset = (methGen.isStatic()  ?  0  :  1);
                                        int numArgs = indeces.length;
                                        modified = true;

                                        // get new signature, update INVOKESTATIC instruction
                                        MethodGen newMethodGen = (MethodGen)newMethodTable.get(methodKey);
                                        invoke_instr = ifactory.createInvoke(className, methodName,
                                                newMethodGen.getReturnType(), newMethodGen.getArgumentTypes(),
                                                de.fub.bytecode.Constants.INVOKESTATIC);
                                        ihandle.setInstruction(invoke_instr);


                                        // swap last few instructions to match order in table
                                        ArrayList oldInstrHandles = new ArrayList();
                                        InstructionHandle ihtemp = ihandle;
                                        for (int i = 0;  i < numArgs;  i++)
                                                oldInstrHandles.add(0, ihtemp = ihtemp.getPrev());

                                        // if (DEBUG)  System.out.println("  old instrh order = " + oldInstrHandles);

                                        // check to see if any of these instructions isn't
                                        // a straight up PUSH of some sort
                                        boolean simplePush = true;
                                        for (int i = 0;  i < numArgs;  i++)
                                        {
                                                InstructionHandle ihcurr = (InstructionHandle)oldInstrHandles.get(i);
                                                Instruction icurr = ihcurr.getInstruction();

                                                if (!(icurr instanceof ACONST_NULL
                                                ||    icurr instanceof BIPUSH
                                                ||    icurr instanceof DCONST
                                                ||    icurr instanceof FCONST
                                                ||    icurr instanceof ICONST
                                                ||    icurr instanceof LCONST
                                                ||    icurr instanceof LDC
                                                ||    icurr instanceof LDC_W
                                                ||    icurr instanceof LDC2_W
                                                ||    icurr instanceof LoadInstruction  // ALOAD, DLOAD, FLOAD, ILOAD, LLOAD
                                                ||    icurr instanceof SIPUSH))
                                                {
                                                        // not a simple push; have to be trickier
                                                        simplePush = false;
                                                        break;
                                                }
                                        }

                                        if (simplePush)
                                        {
                                                if (DEBUG2)  System.out.println("old instrs: " + oldInstrHandles);

                                                // all preceding instructions are just pushes;
                                                // can just reorder them in the method
                                                ArrayList newInstructions = new ArrayList();
                                                for (int i = 0;  i < numArgs;  i++)
                                                {
                                                        int newIndex = indeces[i];  // which var should go i'th

                                                        if (DEBUG2)  System.out.println("  easy way; " + i + " maps to " + newIndex);

                                                        InstructionHandle ih = (InstructionHandle)oldInstrHandles.get(newIndex);
                                                        Instruction icurr = ih.getInstruction();
                                                        if (DEBUG2)  System.out.println("  easy way; icurr = " + icurr);
                                                        newInstructions.add(icurr);
                                                }

                                                // if (DEBUG)  System.out.println("  new instr order = " + newInstructions);

                                                // replace instructions
                                                for (int i = 0;  i < numArgs;  i++)
                                                {
                                                        InstructionHandle ih = (InstructionHandle)oldInstrHandles.get(i);
                                                        Instruction newInstr = (Instruction)newInstructions.get(i);
                                                        ih.setInstruction(newInstr);
                                                }
                                        }
                                        else
                                        {
                                                // need to put into temporary array and pull back out
                                                if (DUP_DEBUG)  System.out.println("Not a simple push situation!");

                                                ArrayList newInstrs = new ArrayList();

            /*
               if (method was one that we changed)
                 make new array of object[]
                 ---
                   PUSH numargs
                   ANEWARRAY 1           args -> argi -> array

                 for each arg i at index n that got reordered to index m,
                   put in slot [m] of array:
                 ---
                   DUP_X1                args -> array -> argi -> array
                   PUSH NEW!index        args -> array -> argi -> array -> j
                   DUP2_X1               args -> array -> array -> j -> argi -> array -> j
                   POP2                  args -> array -> array -> j ->argi
                   AASTORE               args -> array

                 for each index i in array,
                   push array[i] onto stack:

                   DUP                   array -> array
                   PUSH index            array -> array -> i
                   AALOAD                array -> array[i]
                   SWAP                  array[i] -> array


                 POP  // get array off stack

                 invoke
            */

            // varNumTable holds hashtable below
            // hash in [old array index] ==> [new array index] table
                                                //                                      arrayIndexTable.put(new Integer(oldVarList.indexOf(var)), new Integer(i));
                                                Type[] oldArgTypes = (Type[])oldArgTypesTable.get(methodKey);
                                                Type[] newArgTypes = (Type[])newArgTypesTable.get(methodKey);

                                                newInstrs.add(createPushInstruction(oldArgTypes.length));
                                                newInstrs.add(ifactory.createNewArray(de.fub.bytecode.generic.Type.OBJECT, (short)1));

                                                for (int typei = numArgs - 1; typei >= 0; typei--)
                                                {
                                                        // typei is old argument index, must get new remapped index
                                                        // I am assuming no non-reference parameters
                                                        int newIndex = newIndeces[typei];  // indeces[typei];

                                                        if (DEBUG2)  System.out.println("  hard way; " + typei + " maps to " + newIndex);

                                                        newInstrs.add(InstructionConstants.DUP_X1);
                                                        newInstrs.add(createPushInstruction(newIndex));
                                                        newInstrs.add(InstructionConstants.DUP2_X1);
                                                        newInstrs.add(InstructionConstants.POP2);
                                                        newInstrs.add(InstructionFactory.createArrayStore(Type.OBJECT));
                                                }

                                                // now all vars have been put into a temporary array, so
                                                // pull them back out in order
                                                for (int ii = 0; ii < numArgs;  ii++)
                                                {
                                                        newInstrs.add(de.fub.bytecode.generic.InstructionConstants.DUP);
                                                        newInstrs.add(createPushInstruction(ii));
                                                        newInstrs.add(InstructionFactory.createArrayLoad(de.fub.bytecode.generic.Type.OBJECT));
                                                        newInstrs.add(ifactory.createCast(de.fub.bytecode.generic.Type.OBJECT, newArgTypes[ii]));
                                                        newInstrs.add(de.fub.bytecode.generic.InstructionConstants.SWAP);
                                                }

                                                // still need to invoke the static method!
                                                newInstrs.add(InstructionConstants.POP);
                                                newInstrs.add(invoke_instr);

                                                // put the new instructions in the instruction list
                                                int numInstrs = newInstrs.size();
                                                if (numInstrs >  0)
                                                {
                                                        ihandle.setInstruction((de.fub.bytecode.generic.Instruction)newInstrs.get(0));
                                                        for (int ii = 1; ii < numInstrs; ii++)
                                                        {
                                                                Instruction currInstr = (Instruction)newInstrs.get(ii);
                                                                instrList.append(ihandle, currInstr);
                                                                ihandle = ihandle.getNext();
                                                        }
                                                }
                                        }

                                        // if (DEBUG)  System.out.println("  new instructions: " + instrList);
                                        methGen.setInstructionList(instrList);
                                        methGen.setMaxStack();
                                        methGen.setMaxLocals();
                                }

                                Method newMeth = methGen.getMethod();
                                if (DUP_DEBUG)  System.out.println("\nCALL PATCH PHASE, REPLACE METHOD:\n" + meth + " ,\n" + newMeth);
                                cgen.replaceMethod(meth, newMeth);
                        }

                        // done shuffling all methods in this class;
                        // now replace old class with new one
                        if (modified)
                        {
                                JavaClass jc = cgen.getJavaClass();
                                if (DUP_DEBUG)  System.out.println("CFC ADD: " + jc.getClassName());
                                cfc.addClass(jc);
                        }
                }
        }


     
}
