package algorithms.Method2RMadness;





import java.util.*;

import utils.ClassFileCollection;
import utils.ClassHierarchy;
import utils.ConfigProperties;
import utils.MethodID;

import de.fub.bytecode.classfile.*;
import de.fub.bytecode.generic.*;


/** Bludgeons the signatures of methods so that they take Object[] and return Object
  * in order to hide the information implicitly encapsulated in methods' argument lists.
  *
  * <p>
  * Used by Method2RMadness obfuscator.
  *
  * <p>
  * Code quality: Seems to work, but has not been extensively tested.  Main concern:
  * correctness of code by which it is determined whether or not a method is
  * renamable (Sandmark's ClassHierarchy class).
  *
  * @author Martin Stepp (<a href="mailto:stepp@cs.arizona.edu">stepp@cs.arizona.edu</a>)
  * @version 1.0, May 8, 2002
  */
public class SignatureBludgeoner  implements de.fub.bytecode.Constants
{

        /* Turn to true to print debugging messages. */
        private static final boolean DEBUG = false;
        private static final boolean CLASS_PRINT_DEBUG = false;

        /* Constants used for types of instructions. */
        private static final int LOAD = 0, STORE = 1;

        private static final String NEW_METHOD_SIGNATURE = "([Ljava/lang/Object;)Ljava/lang/Object;";
        private static final String[] NEW_ARG_NAMES = new String[] {"args"};
        private static final Type[] NEW_ARG_TYPES   = new Type[] {new ArrayType(Type.OBJECT, 1)};
        private static final Type NEW_RETURN_TYPE   = Type.OBJECT;


        /* Required for SandMark's interface. */
        private ConfigProperties myProps = new ConfigProperties();


        /* Runs a quick test of this obfuscator. */
        public static void main(String[] args) throws Exception
        {
                if (args.length < 1)
                {
                        System.out.println("Usage: SignatureBludgeoner JAR_FILE");
                        System.exit(1);
                }

                try
                {
                        ClassFileCollection cfc = new ClassFileCollection(args[0]);
                        SignatureBludgeoner bludgeoner = new SignatureBludgeoner();
                        bludgeoner.apply(cfc);
                        if (args.length == 2  &&  args[1].equals("-verbose"))
                                printCFC(cfc);
                        cfc.saveJar("SIGBLUD_HARDCODE.jar");
                }
                catch (java.io.IOException e)
                {
                        System.err.println("I/O Error: " + e.getMessage());
                }
        }


        /* To be implemented; returns whether or not we're allowed to reorder the arguments in this method. */
        private static boolean canBludgeon(MethodGen methGen)
        {
                if (methGen == null)
                        return false;

                // only operate on static methods
                if (!methGen.isStatic()  ||  methGen.isSynchronized()  ||  methGen.isNative())
                        return false;

                String methodName = methGen.getName();
                // kludge
                if (methodName.equals("main")  ||  methodName.equals("<clinit>"))
                {
                        if (DEBUG)  System.out.println("special method main or class initializer!  not renaming.");
                        return false;
                }

                return true;
        }

        /* To be implemented; returns whether or not we're allowed to reorder the arguments in this method. */
        private static boolean canBludgeon(JavaClass jc, ClassHierarchy ch, MethodGen methGen)
        {
                if (!canBludgeon(methGen))
                        return false;

                // don't rename methods that Sandmark flags as being non-renamable
                String methodName = methGen.getName();
                String className  = jc.getClassName();
                String methodSig  = methGen.getSignature();

                MethodID methodID   = new MethodID(methodName, methodSig, className);
                MethodID renameToID = new MethodID(methodName, NEW_METHOD_SIGNATURE, className);

                boolean okToRename = false;
                int numOverloads = -1;
                try {
                    okToRename = ch.methodRenameOK(methodID, renameToID);
                    if (okToRename)
                    {
                        MethodID[] overloads = ch.allOverloads(methodID);
                        if (overloads != null)
                                numOverloads = overloads.length;
                    }
                }catch(utils.ClassHierarchyException exc) {
                   // CC
                }
                if (DEBUG)
                {
                        System.out.println("old method: " + methodID);
                        System.out.println("new method: " + renameToID);
                        System.out.println("rename ok?: " + okToRename);
                        System.out.println("overloads : " + numOverloads);
                }

                if (!okToRename)
                        return false;

                // *** future work: make it rename overloads to a unique name
                if (numOverloads >= 1)
                {
                        if (DEBUG)  System.out.println("  Method " + methGen + " has " + numOverloads + " overloads!");
                        return false;
                }

                // make sure method's args are all reference types (Objects, not primitives)
                Type[] varTypes = methGen.getArgumentTypes();
                for (int i = 0;  i < varTypes.length;  i++)
                {
                        if (!(varTypes[i] instanceof ReferenceType))
                                return false;
                }

                // make sure method returns either void or an object
                Type returnType = methGen.getReturnType();
                if (!(returnType == Type.VOID  ||  returnType instanceof ReferenceType))
                        return false;

                return true;
        }


        /* Makes an ICONST_n or BIPUSH n for a given int. */
        private static Instruction createPushInstruction(int n)
        {
                switch (n)
                {
                        case 0:
                        case 1:
                        case 2:
                        case 3:
                        case 4:
                        case 5:
                                return new ICONST(n);
                        default:
                                return new BIPUSH((byte)n);
                }
        }


        /** Prints the contents of the given class file collection to System.out. */
        private static void printCFC(ClassFileCollection cfc)
                throws Exception
        {
                Iterator itr = cfc.classes();
                while (itr.hasNext())
                {
                        String className = (String)itr.next();
                        JavaClass jc = cfc.getClass(className);
                        printClass(jc);
                }
        }


        private static void printClass(JavaClass jc)
        {
                ClassGen cgen = new ClassGen(jc);
                ConstantPoolGen cpgen = cgen.getConstantPool();

                String className = jc.getClassName();
                System.out.println("class " + className + "\n{");

                // shuffle all methods in this class
                int numMethods = cgen.getMethods().length;
                for (int methodNum = 0;  methodNum < numMethods;  methodNum++)
                {
                        // reorder each method
                        Method meth = cgen.getMethodAt(methodNum);
                        MethodGen methGen  = new MethodGen(meth, className, cpgen);
                        System.out.println("  " + meth);
                        // System.out.println(methGen.getInstructionList() + "\n");
                }
                System.out.println("}\n");
        }


        /* Returns a hash key for a method. */
        private static String getMethodKey(JavaClass jc, Method meth)
        {
                return getMethodKey(jc.getClassName(), meth.getName(), meth.getSignature());
        }


        /* Returns a hash key for a method. */
        private static String getMethodKey(String className, String methodName, String methodSignature)
        {
                return className + ":" + methodName + methodSignature;
        }


        /** Bludgeons the method arguments in this ClassFileCollection's classes. */
        public void apply(ClassFileCollection cfc) throws Exception
        {
                // create a class hierarchy for later seeing if methods are renamable
                ClassHierarchy ch = new ClassHierarchy(cfc, myProps);

                ArrayList classNameList = new ArrayList();
                ArrayList classList = new ArrayList();

                Iterator itr = cfc.classes();
                while (itr.hasNext())
                        classNameList.add((String)itr.next());

                Hashtable oldArgTypesTable    = new Hashtable();
                Hashtable oldReturnTypesTable = new Hashtable();
                Hashtable voidMethodTable     = new Hashtable();

                int numClasses = classNameList.size();
                for (int classNum = 0;  classNum < numClasses;  classNum++)
                {
                        String className = (String)classNameList.get(classNum);
                        JavaClass jc = cfc.getClass(className);

                        if (CLASS_PRINT_DEBUG)  System.out.println("Old class:");
                        if (CLASS_PRINT_DEBUG)  printClass(jc);

                        ClassGen cgen = new ClassGen(jc);
                        ConstantPoolGen cpgen = cgen.getConstantPool();
                        InstructionFactory ifactory = new InstructionFactory(cgen, cpgen);

                        // bludgeon all methods in this class
                        Method[] methods = cgen.getMethods();
                        int numMethods = methods.length;
                        for (int methodNum = 0;  methodNum < numMethods;  methodNum++)
                        {
                                // reorder each method
                                Method meth = methods[methodNum];
                                MethodGen methGen  = new MethodGen(meth, className, cpgen);
                                if (!canBludgeon(jc, ch, methGen))
                                {
                                        if (DEBUG)  System.out.println("Can't bludgeon (2): " + methGen);
                                        continue;
                                }
                                if (DEBUG)  System.out.println("Old method: " + meth);

                                // gather up old local variables
                                Type[] oldArgTypes = methGen.getArgumentTypes();
                                int numArgs = oldArgTypes.length;

                                Type oldReturnType = methGen.getReturnType();
                                if (oldReturnType == Type.VOID)
                                {
                                        if (DEBUG)  System.out.println("  Void!  " + meth);
                                        voidMethodTable.put(getMethodKey(jc, meth), "1");  // value not important -- only existence of method as key
                                }

                                // replace them with a single Object[]
                                methGen.setArgumentNames(NEW_ARG_NAMES);
                                methGen.setArgumentTypes(NEW_ARG_TYPES);
                                methGen.setReturnType(NEW_RETURN_TYPE);

                                // fix instructions so they refer to the now-arrayed local variable argument
                                // this involves updating any local variable instruction (a load, store, or int increment)
                                // to make it use the new local variable array.
                                if (DEBUG)  System.out.println("patching instructions in method:\n---");

                                if (numArgs > 0  ||  oldReturnType == Type.VOID)
                                {
                                        InstructionList instrList = methGen.getInstructionList();
                                        if (instrList != null)
                                        {
                                                loop:  for (InstructionHandle ihandle = instrList.getStart();
                                                                                                ihandle != null;
                                                                                                ihandle = ihandle.getNext())
                                                {
                                                        Instruction instr = ihandle.getInstruction();
                                                        if (instr instanceof ALOAD  ||  instr instanceof ASTORE)
                                                        {
                                                                // we have to modify instructions that refer to local vars (arguments)
                                                                LocalVariableInstruction lvinstr = (LocalVariableInstruction)instr;
                                                                if (DEBUG)  System.out.println("L.V. instr: " + instr);
                                                                int oldNum = lvinstr.getIndex();

                                                                // figure out what number of argument this instruction referred to
                                                                if (oldNum >= numArgs)
                                                                {
                                                                        // this instruction refers to a non-argument local variable;
                                                                        // we don't need to modify it
                                                                        if (DEBUG)  System.out.println("  Skipping; number out of range: " + oldNum);
                                                                        continue;
                                                                }

                                                                // build new instructions that refer correctly to new local array variable
                                                                // (instead of an ALOAD,  load the array, push the index, then do AALOAD;
                                                                //  instead of an ASTORE, load the array, push the index, then do AASTORE)
                                                                ArrayList newInstrs = new ArrayList();
                                                                newInstrs.add(InstructionFactory.createLoad(Type.OBJECT, 0));  // aload_x
                                                                newInstrs.add(createPushInstruction(oldNum));
                                                                if (instr instanceof ALOAD)
                                                                {
                                                                        newInstrs.add(InstructionFactory.createArrayLoad(Type.OBJECT));
                                                                        newInstrs.add(ifactory.createCast(Type.OBJECT, oldArgTypes[oldNum]));
                                                                }
                                                                else if (instr instanceof ASTORE)
                                                                {
                                                                        newInstrs.add(new DUP2_X1());
                                                                        newInstrs.add(new POP2());
                                                                        newInstrs.add(InstructionFactory.createArrayStore(Type.OBJECT));
                                                                }

                                                                if (DEBUG)  System.out.println("  new instrs = " + newInstrs);

                                                                // update method's instruction list to replace old instruction with new ones
                                                                int newInstrsSize = newInstrs.size();

                                                                // replace original one with last one from new list; this keeps us
                                                                // from having to worry about "lost targets" and other such filth
                                                                if (newInstrsSize > 0)
                                                                {
                                                                        ihandle.setInstruction((Instruction)newInstrs.get(0));
                                                                        for (int i = 1;  i < newInstrsSize;  i++)
                                                                        {
                                                                                instrList.append(ihandle, (Instruction)newInstrs.get(i));
                                                                                ihandle = ihandle.getNext();
                                                                        }
                                                                }
                                                        }
                                                        else if (instr instanceof RETURN)
                                                        {
                                                                // insert a push of null just before return, and areturn it
                                                                ihandle.setInstruction(InstructionConstants.ACONST_NULL);
                                                                instrList.append(ihandle, InstructionConstants.ARETURN);
                                                                ihandle = ihandle.getNext();
                                                        }
                                                }
                                        }

                                        methGen.setInstructionList(instrList);

                                        // replace method in class
                                        if (cgen.containsMethod(methGen.getName(), methGen.getSignature()) != null)
                                        {
                                                /*
                                                System.out.println();
                                                System.out.println();
                                                printClass(cgen.getJavaClass());
                                                throw new RuntimeException("Already have this method: " + methGen.getMethod());
                                                */
                                        }
                                }

                                // finish up
                                methGen.setMaxStack();
                                methGen.setMaxLocals();

                                Method newMeth = methGen.getMethod();
                                if (DEBUG)  System.out.println("\nNew method: " + newMeth);

                                cgen.replaceMethod(meth, newMeth);

                                // store the hash table of old ==> new var #s, so we can fix calls later
                                oldArgTypesTable.put(getMethodKey(jc, meth), oldArgTypes);
                                oldReturnTypesTable.put(getMethodKey(jc, meth), oldReturnType);
                        }

                        // done shuffling all methods in this class;
                        // now replace old class with new one
                        JavaClass newClass = cgen.getJavaClass();
                        cfc.addClass(newClass);
                        classList.add(cgen);
                }



                // fix all calls by making them build an Object[]
                // (also clean up void methods by popping stack)
                if (DEBUG)  System.out.println("\nfixing all calls to modified methods:\n---");

                for (int classNum = 0;  classNum < numClasses;  classNum++)
                {
                        String callingClassName = (String)classNameList.get(classNum);
                        ClassGen cgen = (ClassGen)classList.get(classNum);
                        ConstantPoolGen cpgen = cgen.getConstantPool();
                        InstructionFactory ifactory = new InstructionFactory(cgen, cpgen);

                        boolean modified = false;

                        int numMethods = cgen.getMethods().length;
                        for (int methodNum = 0;  methodNum < numMethods;  methodNum++)
                        {
                                Method meth = cgen.getMethodAt(methodNum);
                                MethodGen methGen  = new MethodGen(meth, callingClassName, cpgen);
                                InstructionList instrList = methGen.getInstructionList();
                                if (instrList == null)
                                        continue;

                                for (InstructionHandle ihandle = instrList.getStart();
                                     ihandle != null;
                                     ihandle = ihandle.getNext())
                                {
                                        Instruction instr = ihandle.getInstruction();
                                        if (!(instr instanceof INVOKESTATIC))
                                                continue;

                                        // patch invokestatic calls so that they push object array
                                        // instead of individual arguments
                                        INVOKESTATIC invoke_instr = (INVOKESTATIC)instr;
                                        if (DEBUG)  System.out.println("in " + callingClassName + ":" + meth + " at " + instr + ":");

                                        String className  = invoke_instr.getClassName(cpgen);
                                        String methodName = invoke_instr.getMethodName(cpgen);
                                        String methodSig  = invoke_instr.getSignature(cpgen);
                                        String methodKey  = getMethodKey(className, methodName, methodSig);

                                        if (DEBUG)  System.out.println("  className = " + className + ", methodName = " + methodName + ", methodSig = " + methodSig);
                                        if (!oldArgTypesTable.containsKey(methodKey))
                                        {
                                                if (DEBUG)  System.out.println("  Skipping; can't find variable reordering table!");
                                                continue;
                                        }

                                        // if we get this far, we know this is a method we obfuscated,
                                        // so fix the call
                                        modified = true;
                                        Type[] oldArgTypes = (Type[])oldArgTypesTable.get(methodKey);
                                        Type oldReturnType = (Type)oldReturnTypesTable.get(methodKey);
                                        int numArgs = oldArgTypes.length;

                                        // change instructions that push arguments into one instruction that pushes an array.
                                        ArrayList newInstrs = new ArrayList();

                                        // bipush numArgs
                                        // anewarray Object 1-dimension
                                        newInstrs.add(createPushInstruction(numArgs));
                                        newInstrs.add(ifactory.createNewArray(Type.OBJECT, (short)1));

                                        /* for each arg i on stack (inverse order),
                                           start    dup     dup2_x1,pop2   bipush i   swap   aastore
                                        */
                                        for (int i = numArgs - 1;  i >= 0;  i--)
                                        {
                                                newInstrs.add(InstructionConstants.DUP);
                                                newInstrs.add(InstructionConstants.DUP2_X1);
                                                newInstrs.add(InstructionConstants.POP2);
                                                newInstrs.add(createPushInstruction(i));
                                                newInstrs.add(InstructionConstants.SWAP);
                                                newInstrs.add(InstructionFactory.createArrayStore(Type.OBJECT));
                                        }

                                        // now call new method
                                        newInstrs.add(ifactory.createInvoke(className, methodName,
                                                NEW_RETURN_TYPE, NEW_ARG_TYPES, INVOKESTATIC));

                                        // now pop null object off stack if old method was void
                                        if (voidMethodTable.containsKey(methodKey))
                                                newInstrs.add(InstructionConstants.POP);
                                        else
                                                newInstrs.add(ifactory.createCast(Type.OBJECT, oldReturnType));

                                        if (DEBUG)  System.out.println("  newInstrs = " + newInstrs);

                                        // put in new instructions
                                        int newInstrsSize = newInstrs.size();
                                        if (newInstrsSize > 0)
                                        {
                                                ihandle.setInstruction((Instruction)newInstrs.get(0));

                                                for (int i = 1;  i < newInstrsSize;  i++)
                                                {
                                                        instrList.append(ihandle, (Instruction)newInstrs.get(i));
                                                        ihandle = ihandle.getNext();
                                                }
                                        }

                                        if (DEBUG)  System.out.println("  new instructions: " + instrList);
                                }

                                methGen.setInstructionList(instrList);
                                methGen.setMaxStack();
                                methGen.setMaxLocals();

                                Method newMeth = methGen.getMethod();
                                cgen.replaceMethod(meth, newMeth);
                        }

                        // done shuffling all methods in this class;
                        // now replace old class with new one

                        if (modified)
                        {
                                JavaClass newClass = cgen.getJavaClass();
                                cfc.addClass(newClass);

                                if (CLASS_PRINT_DEBUG)  System.out.println("New class: ");
                                if (CLASS_PRINT_DEBUG)  printClass(newClass);
                        }
                }
        }


      
     
}
